/*
 *   Copyright (c) 1994, 1998 Hellmuth Michaelis. All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *   
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *   SUCH DAMAGE.
 *
 *---------------------------------------------------------------------------
 *
 *	tina_dump - i4b Stollman Tina-dd dual port memory dumper
 *	--------------------------------------------------------
 *
 *	$Id: tina_dump.c,v 1.3 1999/05/25 09:10:41 hm Exp $
 *
 *	last edit-date: [Tue May 25 11:14:07 1999]
 *
 *---------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/file.h>
#include <sys/ioctl.h>

#include "i4b_tina_ioctl.h"

#define COLS	16
#define LINES	0x100

#define DEFAULT_DEVICE	"/dev/tina0"

int fd;

int do_isdn = 0;
int do_addr = 0;
int do_leng = 0;

unsigned int addr;
int leng;

void usage ( void );

int main(int argc, char *argv[])
{
	extern int optind;
	extern int opterr;
	extern char *optarg;

	char *isdnname = "";
	unsigned char *rdbuffer;	
	int c;
	int i;
	struct record record;
	int row, col, curcol, savecol, extracol;

	while( (c = getopt(argc, argv, "a:l:i:")) != -1)
	{
		switch(c)
		{
			case 'i':
				isdnname = optarg;
				do_isdn = 1;
				break;

			case 'a':
				sscanf(optarg, "%x", &addr);
				do_addr = 1;
				break;
				
			case 'l':
				sscanf(optarg, "%d", &leng);
				do_leng = 1;
				break;
				
			case '?':
			default:
				usage();
				break;
		}
	}

	if(!do_isdn)
		isdnname = DEFAULT_DEVICE;
	
	if((fd = open(isdnname, O_RDONLY)) == -1)
	{
		char buffer[80];
		strcpy(buffer, "cannot open isdn devicefile: ");
		strcat(buffer, isdnname);
		perror(buffer);
		exit(1);
	}

	if(!(do_addr && do_leng))
		usage();

	if((rdbuffer = malloc(leng+1)) == NULL)
	{
		fprintf(stderr, "malloc failed to allocate %d bytes, exit!\n", leng+1);
		exit(1);
	}		
	
	record.addr = addr;
	record.length = leng;
	record.data = rdbuffer;
	
	if(ioctl(fd, ISDN_GETBLK, &record) == -1)
	{
		perror("ioctl ISDN_GETBLK failed");
		exit(1);
	}

	i = 0;
	row = 0;

	while(leng > 0)
	{
		if(leng > COLS)
			col = 16;
		else
			col = leng;
			
		savecol = col;
		curcol = 0;
		extracol = 16 - col;

		if((addr == 0) || (!(addr % LINES)))
		{
			printf("\n\n");
			printf("Adress  0  1  2  3  4  5  6  7   8  9  A  B  C  D  E  F   0123456789ABCDEF\n");
			printf("------  -----------------------  -----------------------  ----------------\n");
			printf("%06.6X  ",addr);			
		}
		else
		{
			printf("\n%06.6X  ",addr);
		}
			
		while(col--)
		{
			printf("%02.2X ", rdbuffer[i+curcol++]);

			if(curcol == (COLS/2))
				printf(" ");
		}
		while(extracol--)
		{
			printf("   ");
			if(curcol++ == (COLS/2))
				printf(" ");
		}

		printf(" ");
			
		curcol = 0;

		while(savecol--)
		{
			if(isprint(rdbuffer[i+curcol]))
			{
				printf("%c", rdbuffer[i+curcol++]);
			}
			else
			{
				curcol++;
				printf(".");
			}
		}
		addr += COLS;
		leng -= COLS;
		i += COLS;
	}
	printf("\n");
	return(0);
}

void usage()
{
	fprintf(stderr,"\ndump - dump tina-dd memory contents to stdout\n");
	fprintf(stderr,"usage: dump [-i <name>] [-a addr ] [-l length]\n");
	fprintf(stderr,"       -i <devname>  isdn devicefile to download to \n");
	fprintf(stderr,"       -a <addr>     start addr in tina-dd memory\n");
	fprintf(stderr,"       -l <length>   length of memory dump in bytes\n");
	fprintf(stderr,"\n");
	exit(1);
}
