/*
 *   Copyright (c) 1994, 1999 Hellmuth Michaelis. All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *   SUCH DAMAGE.
 *
 *---------------------------------------------------------------------------
 *
 *	i4b_tina_api.c - i4b Stollman Tina-dd card API interface
 *	--------------------------------------------------------
 *
 *	$Id: i4b_tina_api.c,v 1.2 1999/05/04 09:07:29 hm Exp $
 *
 *	last edit-date: [Mon May  3 15:45:04 1999]
 *
 *---------------------------------------------------------------------------*/

#include "tina.h"

#if NTINA > 0

#include <sys/param.h>
#if defined(__FreeBSD__) && __FreeBSD__ >= 3
#include <sys/ioccom.h>
#else
#include <sys/ioctl.h>
#endif
#include <sys/kernel.h>
#include <sys/systm.h>
#include <sys/conf.h>
#include <sys/mbuf.h>
#include <sys/proc.h>
#include <sys/fcntl.h>
#include <sys/socket.h>
#include <net/if.h>

#ifdef __FreeBSD__

#include <i386/isa/isa_device.h>

#include <machine/i4b_debug.h>
#include <machine/i4b_ioctl.h>

#else

#include <machine/bus.h>
#include <sys/device.h>

#include <i4b/i4b_global.h>
#include <i4b/i4b_debug.h>
#include <i4b/i4b_ioctl.h>

#endif

#include <i4b/include/i4b_mbuf.h>
#include <i4b/include/i4b_l3l4.h>

#include <i4b/tina-dd/i4b_tina_ioctl.h>
#include <i4b/tina-dd/i4b_tina_api.h>

#ifdef NOTYET

#define MAX_APPLID 2

ccb_t ccb[MAX_APPLID];	/* channel control blocks for each applid */

#endif

/*---------------------------------------------------------------------------*
 *	send a port command to firmware
 *---------------------------------------------------------------------------*/
void
tina_exec_mrin(tina_sc_t *sc, struct MRIN *mrinp)
{
	/* set busy status int status byte */
	
	mrinp->MRSTAT = MRIBSY;

	/* write command to input register */
	
	(void) tina_writeblock(sc->sc_iobase, (u_char *) mrinp,
				MRIN_ADDR, sizeof(struct MRIN));

	/* trigger irq on tina-dd to start execution of command */
	
	outb(((sc->sc_iobase)+CTRL_STAT), CR_SET_INTP);
}

/*---------------------------------------------------------------------------*
 *	get mrint from firmware
 *---------------------------------------------------------------------------*/
void
tina_get_mrint(tina_sc_t *sc, struct MRINT *mrintp)
{
	(void) tina_readblock(sc->sc_iobase, MRINT_ADDR,
			(u_char *) mrintp, sizeof(struct MRINT));
}

/*---------------------------------------------------------------------------*
 *	get mrout from firmware
 *---------------------------------------------------------------------------*/
void
tina_get_mrout(tina_sc_t *sc, struct MRINT *mroutp)
{
	(void) tina_readblock(sc->sc_iobase, MROUT_ADDR,
			(u_char *) mroutp, sizeof(struct MROUT));
}

#ifdef NOTYET

/*---------------------------------------------------------------------------*
 *	open ADMINISTRATION channel to FCAPI
 *---------------------------------------------------------------------------*/
int
admin_ch_open(void)
{
	struct MRIN MRIN;		/* module input register */
	
	bzero((char *)&MRIN, sizeof(struct MRIN));

	MRIN.MRCOD = MR_P_INIT;
	MRIN.PORTNO = 0;		

	tina_exec_mrin(sc, &MRIN))
}

/*---------------------------------------------------------------------------*
 *	open CONTROL cannel to FCAPI
 *---------------------------------------------------------------------------*/
int
control_ch_open(void)
{
	struct MRIN MRIN;		/* module input register */
	struct MROUT *pmrout;		/* module output register */
	u_short channel;		/* control channel number */

	bzero((char *)&MRIN, sizeof(struct MRIN));

	MRIN.MRCOD = MR_GET_CHNL;	/* get a control channel */
	MRIN.PORTNO = 0;		/* from admin channel */
	MRIN.MRP1 = 0x0a;		/* module id, high byte = FCAPI */
	MRIN.MRP2 = 0x00;		/* module id, low byte = FCAPI */	

	if((pmrout = exe_mrin(fd, &MRIN)) == NULL)
		return(-1);

	if(pmrout->MRERR == 0)		/* no error ? */
	{
		channel  = pmrout->MRP2 << 8;	/* high byte of channel no */
		channel |= pmrout->MRP1;	/* low byte of channel no */
	}
	else
	{
		printf("ERROR: CONTROL GET_CHNL command failed!\n");
		return -1;		/* return error */
	}

	bzero((char *)&MRIN, sizeof(struct MRIN));

	MRIN.MRCOD = MR_P_OPEN;		/* open the control channel */
	MRIN.PORTNO = channel;		/* channel number from get_chnl */

	if((pmrout = exe_mrin(fd, &MRIN)) == NULL)
		return(-1);

	if(pmrout->MRERR == 0)		/* no error ? */
	{
		return(channel);
	}
	else
	{
		printf("ERROR: CONTROL P_OPEN command failed!\n");
		return(-1);
	}
}

/*---------------------------------------------------------------------------*
 *	open DATA cannel to FCAPI
 *---------------------------------------------------------------------------*/
int
data_ch_open(int control_ch)
{
	struct MRIN MRIN;		/* module input register */
	struct MROUT *pmrout;		/* module output register */
	u_short channel;		/* control channel number */

	bzero((char *)&MRIN, sizeof(struct MRIN));

	MRIN.MRCOD = MR_GET_CHNL;	/* get a control channel */
	MRIN.PORTNO = (BYTE)control_ch;	/* from related control channel */
	MRIN.MRP1 = 0x0a;		/* module id, high byte = FCAPI */
	MRIN.MRP2 = 0x00;		/* module id, low byte = FCAPI */	

	if((pmrout = exe_mrin(fd, &MRIN)) == NULL)
		return(-1);

	if(pmrout->MRERR == 0)		/* no error ? */
	{
		channel  = pmrout->MRP2 << 8;	/* high byte of channel no */
		channel |= pmrout->MRP1;	/* low byte of channel no */
	}
	else
	{
		printf("ERROR: DATA GET_CHNL command failed!\n");
		return -1;		/* return error */
	}

	bzero((char *)&MRIN, sizeof(struct MRIN));

	MRIN.MRCOD = MR_P_OPEN;		/* open the control channel */
	MRIN.PORTNO = channel;		/* channel number from get_chnl */

	if((pmrout = exe_mrin(fd, &MRIN)) == NULL)
		return(-1);

	if(pmrout->MRERR == 0)		/* no error ? */
	{
		return(channel);
	}
	else
	{
		printf("ERROR: DATA P_OPEN command failed!\n");
		return(-1);
	}
}

/*---------------------------------------------------------------------------*
 *	close a cannel
 *---------------------------------------------------------------------------*/
int
channel_close(u_short channel)
{
	struct MRIN MRIN;		/* module input register */
	struct MROUT *pmrout;		/* module output register */

	bzero((char *)&MRIN, sizeof(struct MRIN));

	MRIN.MRCOD = MR_P_CLOSE;	/* close the control channel */
	MRIN.PORTNO = channel;		/* channel number from get_chnl */

	if((pmrout = exe_mrin(fd, &MRIN)) == NULL)
		return(-1);

	return(0);
}

/*---------------------------------------------------------------------------*
 *	return channel number for given applid
 *---------------------------------------------------------------------------*/
static int get_channel(int appl_id, capi_msg_t *msg)
{
	int i;
	int empty = -1;
	int channel = -1;
	
	/* search for application id */
	
	for(i=0; i < MAX_APPLID; i++)
	{
		if(ccb[i].applid == appl_id)
			break;
		if(empty == -1 && ccb[i].applid == -1)
			empty = i;
	}

	/* not found, get empty one */
	
	if(ccb[i].applid != appl_id)
	{
		i = empty;
		ccb[i].applid = appl_id;
	}

	switch(msg->command)
	{

		/* messages going thru DATA channel */

		case LISTEN_B3_REQ:
		case CONNECT_B3_REQ:
		case CONNECT_B3_RESP:
		case CONNECT_B3_ACTIVE_RESP:
		case DISCONNECT_B3_REQ:
		case DISCONNECT_B3_RESP:
		case GET_B3_PARAMS_REQ:
		case DATA_B3_REQ:
		case DATA_B3_RESP:
		case RESET_B3_REQ:
		case RESET_B3_RESP:
			if(ccb[i].data_ch == -1)
				ccb[i].data_ch = data_ch_open(ccb[i].cntl_ch);
			channel = ccb[i].data_ch;
			break;

		/* messages going thru CONTROL channel */

		default:
			if(ccb[i].cntl_ch == -1)
				ccb[i].cntl_ch = control_ch_open();
			channel = ccb[i].cntl_ch;
			break;
	}
	return(channel);
}

/*---------------------------------------------------------------------------*
 *	reset one ccb entry
 *---------------------------------------------------------------------------*/
static int reset_ccb(int applid)
{
	int i;

	/* search for application id */
	
	for(i=0; i < MAX_APPLID; i++)
	{
		if(ccb[i].applid == applid)
			break;
	}

	/* not found, exit */
	
	if(ccb[i].applid != applid)
		return(-1);

	ccb[i].applid = -1;
	ccb[i].cntl_ch = -1;
	ccb[i].data_ch = -1;

	return(0);
}
	
/*---------------------------------------------------------------------------*
 *	initialize capi interface
 *---------------------------------------------------------------------------*/
int api_init(void)
{
	int i;

	/* init all ccb entries */
	
	for(i=0; i < MAX_APPLID; i++)
	{
		ccb[i].applid = -1;
		ccb[i].cntl_ch = -1;
		ccb[i].data_ch = -1;
	}

	/* open administration channel */
	
	return(admin_ch_open());
}

/*---------------------------------------------------------------------------*
 *	de-initialize capi interface
 *---------------------------------------------------------------------------*/
int api_finish(int applid)
{
	int i;

	/* search for application id */
	
	for(i=0; i < MAX_APPLID; i++)
	{
		if(ccb[i].applid == applid)
			break;
	}

	/* not found, exit */
	
	if(ccb[i].applid != applid)
		return(-1);

	if(ccb[i].data_ch != -1)
		channel_close(ccb[i].data_ch);

	if(ccb[i].cntl_ch != -1)
		channel_close(ccb[i].cntl_ch);

	reset_ccb(ccb[i].applid);
	
	return(0);
}

/*---------------------------------------------------------------------------*
 *	send a CAPI message to the controller
 *---------------------------------------------------------------------------*/
int api_put_message(int appl_id, capi_msg_t *msg)
{

#define RETRY_MAX	30
#define RETRY_SLEEP	1

	struct MRIN MRIN;		/* module input register */
	struct MROUT *pmrout;		/* module output register */
	int channel;			/* channel to use */
	int retry;			/* getbuf retry counter */
	unsigned int buf_addr;		/* seg:off buffer address */
	unsigned int cbuf_addr;		/* converted buffer address */	
	struct record record;		/* writing messages */

	if( (channel = get_channel(appl_id, msg)) == -1)
	{
		printf("ERROR: get_channel() failed!\n");
		return(-1);
	}

	retry = RETRY_MAX;
	
	for(;;)
	{
		bzero((char *)&MRIN, sizeof(struct MRIN));

		MRIN.MRCOD = MR_GETBUF;
		MRIN.PORTNO = channel;		/* channel no */

		MRIN.MRP1 = msg->total_len & 0xff;
		MRIN.MRP2 = (msg->total_len >> 8) & 0xff;
	
		if(((pmrout = exe_mrin(fd, &MRIN)) == NULL) ||
			(pmrout->MRERR == 0x08))
		{
			if(--retry)
			{
				sleep(RETRY_SLEEP);	/* try again */
				continue;
			}
			else
			{
				printf("ERROR: exe_mrin(GETBUF) failed!\n");
				return(-1);	/* error, retry exhausted */
			}
		}
		else
		{
			if(retry != RETRY_MAX)
				printf("exe_mrin: retried GETBUF %d times\n", RETRY_MAX-retry);
			break;	/* no error, done */
		}
	}

	if(pmrout->MRERR == 0)		/* no error ? */
	{
		buf_addr = pmrout->MRP4;
		buf_addr = buf_addr << 8;
		buf_addr |= pmrout->MRP3;
		buf_addr = buf_addr << 8;
		buf_addr |= pmrout->MRP2;
		buf_addr = buf_addr << 8;
		buf_addr |= pmrout->MRP1;
	}
	else
	{
		printf("ERROR: GETBUF command failed with error 0x%04x!\n", pmrout->MRERR);
		return(-1);		/* return error */
	}

	bzero((char *)&MRIN, sizeof(struct MRIN));

	MRIN.MRCOD = MR_WRITE;
	MRIN.PORTNO = channel;			/* channel no */

	MRIN.MRP1 = buf_addr & 0xff;
	MRIN.MRP2 = (buf_addr >> 8) & 0xff;
	MRIN.MRP3 = (buf_addr >> 16) & 0xff;
	MRIN.MRP4 = (buf_addr >> 24) & 0xff;	

	MRIN.MRP5 = msg->total_len & 0xff;
	MRIN.MRP6 = (msg->total_len >> 8) & 0xff;

	cbuf_addr = conv_addr(buf_addr);
	
	record.length = msg->total_len;
	record.addr = cbuf_addr;
	record.data = (unsigned char *)msg;	

	/* write message to buffer */
	
	if(ioctl(fd, ISDN_SETBLK, &record) == -1)
	{
		perror("ioctl ISDN_SETBLK failed");
		return(-1);
	}

	/* execute write command */
	
	if((pmrout = exe_mrin(fd, &MRIN)) == NULL)
	{
		printf("ERROR: exe_mrin(WRITE) failed!\n");
		return(-1);
	}

	return(0);
}

/*---------------------------------------------------------------------------*
 *	get a CAPI message from the controller with WAIT !!!!
 *---------------------------------------------------------------------------*/
int api_get_message_w(int appl_id, capi_msg_t *msg)
{
	struct MRINT MRINT;		/* module irq struct at msg irq time */

	struct message message;

	/* get MRINT */
	
	if(ioctl(fd, ISDN_GETMI_PORTCMD, &MRINT) == -1)
	{
		perror("ioctl ISDN_GETMI_PORTCMD failed");
		return(-1);
	}

	if(debug > 1)
		print_mrint(&MRINT);

	/* get message */

	message.data = (unsigned char *)msg;
	
	if(ioctl(fd, ISDN_GETMM_PORTCMD, &message) == -1)
	{
		perror("ioctl ISDN_GETMM_PORTCMD failed");
		return(-1);
	}
	if(debug > 1)
		print_message(&message);

	if(debug)
		displ_capi_msg(msg);

	return(0);
}

/*---------------------------------------------------------------------------*
 *	get a CAPI message from the controller
 *---------------------------------------------------------------------------*/
int api_get_message(int appl_id, capi_msg_t *msg)
{
	struct message message;

	/* get message */

	message.data = (unsigned char *)msg;
	
	if(ioctl(fd, ISDN_GETMM_PORTCMD, &message) == -1)
	{
		perror("ioctl ISDN_GETMM_PORTCMD failed");
		return(-1);
	}
	if(debug > 1)
		print_message(&message);

	if(debug)
		displ_capi_msg(msg);

	return(0);
}

#endif /* NOTYET */

#endif /* NTINA > 0 */

/* EOF */
