/*
 *   Copyright (c) 1994, 1999 Hellmuth Michaelis. All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *   SUCH DAMAGE.
 *
 *---------------------------------------------------------------------------
 *
 *	i4b_tina_api.h - i4b Stollman Tina-dd card API interface
 *	--------------------------------------------------------
 *
 *	$Id: i4b_tina_api.h,v 1.3 1999/05/04 13:51:53 hm Exp $
 *
 *	last edit-date: [Tue May  4 14:36:56 1999]
 *
 *---------------------------------------------------------------------------*/

#ifndef _I4B_TINA_API_H_
#define _I4B_TINA_API_H_

#define MAX_APPLID 2

#ifdef NOTDEF

typedef struct {
	int applid;	/* application id */
	int cntl_ch;	/* fcapi control channel */
	int data_ch;	/* fcapi data channel */
} ccb_t;

int admin_ch_open(void);
int control_ch_open(void);
int data_ch_open(int control_ch);
int channel_close(u_short channel);

int api_init(void);
int api_finish(int applid);
int api_put_message(int appl_id, capi_msg_t *msg);
int api_get_message_w(int appl_id, capi_msg_t *msg);
int api_get_message(int appl_id, capi_msg_t *msg);

#endif

typedef struct {
	int sc_unit;
	int sc_iobase;
	int sc_state;	
} tina_sc_t;

extern tina_sc_t tina_sc[];

/*---------------------------------------------------------------------------*
 *	module register message
 *---------------------------------------------------------------------------*/

#define MAX_MSGLEN (2048+128)		/* CAPI Data + FCAPI module overhead */

struct message {
	unsigned int length;		/* length of data block		   */
	unsigned char *data;		/* pointer to the datablock itself */
};

/*---------------------------------------------------------------------------*
 *	module register extension	(not used for FCAPI)
 *---------------------------------------------------------------------------*/

#define MRCTL_DATA_LEN 60			/* data length max = 60 byte */

struct	MRCTL {
	unsigned short CTLLEN;			/* length of data */
	unsigned short PROTID;			/* reserved */
	unsigned char data[MRCTL_DATA_LEN];	/* data */
};

/*---------------------------------------------------------------------------*
 *	MRIN, structure used by host to send commands to tina-dd
 *---------------------------------------------------------------------------*/

#define	MRIN_ADDR	0x300		/* MRIN addr in dual ported mem */

struct MRIN {
	unsigned char  MRSTAT;		/* status byte		*/
#define		MRIBSY	0x80		/* 	register busy	*/
#define 	MRPOLL	0x01		/*	no irq generated*/

	unsigned char  MRCOD;		/* command byte		*/
	unsigned short reserved;	/*             reserved */
	unsigned short PORTNO;		/* channel number	*/
	unsigned char  MRP1;		/* parameter byte 1	*/
	unsigned char  MRP2;		/* parameter byte 2	*/
	unsigned char  MRP3;		/* parameter byte 3	*/
	unsigned char  MRP4;		/* parameter byte 4	*/
	unsigned char  MRP5;		/* parameter byte 5	*/
	unsigned char  MRP6;		/* parameter byte 6	*/
	struct MRCTL MRCTL;		/* control field	*/
};

/*---------------------------------------------------------------------------*
 *	MROUT, structure used by tina-dd to send results to host
 *---------------------------------------------------------------------------*/

#define	MROUT_ADDR	0x34c		/* MROUT addr in dual ported mem */

struct MROUT {
	unsigned char  MRSTAT;		/* status byte		           */
#define 	MROBSY	0x40		/* 	command not implemented	   */
#define 	MRNEG	0x20		/*	error occured if 1	   */

	unsigned char  MRCOD;		/* command byte	(copied from MRIN) */
	unsigned short MRERR;		/* error code if MRSTAT.MRNEG is 1 */
	unsigned char  MRP1;		/* parameter byte 1	*/
	unsigned char  MRP2;		/* parameter byte 2	*/
	unsigned char  MRP3;		/* parameter byte 3	*/
	unsigned char  MRP4;		/* parameter byte 4	*/
	unsigned char  MRP5;		/* parameter byte 5	*/
	unsigned char  MRP6;		/* parameter byte 6	*/
	unsigned char  MRP7;		/* parameter byte 7	*/
	unsigned char  MRP8;		/* parameter byte 8	*/	
	struct MRCTL MRCTL;		/* control field	*/
};

/*---------------------------------------------------------------------------*
 *	MRINT, structure used by tina-dd to send results to host at IRQ time
 *---------------------------------------------------------------------------*/

#define	MRINT_ADDR	0x398		/* MRINT addr in dual ported mem */

struct MRINT {
	unsigned char  MRISTAT;		/* status byte		            */
#define 	MRIBSY	0x80		/* 	register busy               */

	unsigned char  MRTYPD;		/* type of interrupt                */
#define		MRIRTR	1		/*	channel ready to receive    */
#define 	MRIPEN	2		/*	message pending             */
					/*	3 = reserved                */
#define 	MRIERR	4		/*	B-channel protocol error    */
#define 	MRIRDY	5		/*	channel is ready            */
#define 	MRIECD	6		/*	error executing MRIN cmd    */
#define 	MRIOUT	7		/*	MROUT ready, command exec'd */

	unsigned short MRIPRT;		/* channel number	*/
	unsigned short MRIRCD;		/* error code, reserved	*/
	unsigned short MRIMLN;		/* message length	*/
	unsigned char MSGA1;		/* message addr if MRTYPD = MRIPEN */
	unsigned char MSGA2;		/* message addr if MRTYPD = MRIPEN */
	unsigned char MSGA3;		/* message addr if MRTYPD = MRIPEN */
	unsigned char MSGA4;		/* message addr if MRTYPD = MRIPEN */	
	struct MRCTL MRCTL;		/* control field	*/
};

/*---------------------------------------------------------------------------*
 *	valid commands for module communication
 *---------------------------------------------------------------------------*/

#define	MR_GETBUF	0x01		/* get onboard buffer */
#define	MR_GETXBUF	0x81		/* get expedited onboard buffer */
#define MR_RELBUF	0x02		/* release onboard buffer */
					/* 0x03 reserved */
#define MR_WRITE	0x04		/* send data to module */
#define MR_READ		0x05		/* get data from module */
					/* 0x06 reserved */
#define MR_RESCON	0x07		/* ?? - reserved */
#define MR_P_OPEN	0x08		/* open channel to module */
#define MR_P_CLOSE	0x09		/* close channel to module */
#define MR_P_INIT	0x0a		/* initialize module */
#define MR_FLUSHBUF	0x0b		/* release all buffers for channel */
					/* 0x0c reserved */
					/* 0x0d reserved */
					/* 0x0e reserved */
#define MR_GET_CHNL	0x0f		/* get channel for board communication */
#define MR_P_MODMEM	0x10		/* ?? - reserved */
#define MR_P_MODSTART	0x11		/* ?? - reserved */
#define MR_P_MODSTOP	0x12		/* ?? - reserved */

/* addresses have to be specified in segment and offset .. :-( */

#define FP_OFF(fp)	((unsigned)(fp))
#define FP_SEG(fp)	((unsigned)((unsigned long)(fp) >> 16))

/* forward declarations */

void tina_setupaddr(u_short iobase, u_int addr);
u_int tina_getaddr(u_short iobase);

u_char tina_readbyte(u_short iobase, u_int addr);
void tina_writebyte(u_short iobase, u_int addr, u_char byte);

void tina_readblock(u_short iobase, u_int addr, u_char *dst, u_int len);
void tina_writeblock(u_short iobase, u_char *src, u_int addr, int len);

void tina_exec_mrin(tina_sc_t *sc, struct MRIN *mrinp);
void tina_get_mrint(tina_sc_t *sc, struct MRINT *mrintp);
void tina_get_mrout(tina_sc_t *sc, struct MRINT *mroutp);

#endif /* _I4B_TINA_API_H_ */
